//--------------------------------------------------------------------------------
// PS2 DMA related functionality
// Low-level dma functions.  For setting up and building dma lists.
//
// Geoff Audy, October 26th 2001  Initial functionality layout
// Geoff Audy, May 8th 2002       Change DMA completion polling to good one
// Geoff Audy, May 10th 2002      Fixed a bug where 'DMA_ChainAdd' wouldn't update
//                                the 'spaceleft' element when adding inline data
//                                (so it didn't always know it overflowed)
// Geoff Audy, July 10th 2002     Modified to run under PS2Linux
//
// Copyright (C)2002 Sony Computer Entertainment America Inc.
//
// This file is subject to the terms and conditions of the GNU
// General Public Licence Version 2.  See the file "COPYING"
// in the main directory of this archive for more details.
//--------------------------------------------------------------------------------

#include <stdlib.h>              // Standard C Headers
#include <stdio.h>               //

#include <string.h>              //
#include <ps2gs.h>               //
#include <ps2dma.h>              //
#include <ps2gs.h>               //
#include <ps2vpu.h>              //
#include <ps2vpufile.h>          //

#include "os.h"                  // Project Headers
#include "ps2_dma.h"             //

//-------------------------------------
#define  DMA_POWEROF2(_x)  (!((_x) & ((_x) - 1)))

enum
{
   DMA_ID_REFE = 0,
   DMA_ID_CNT,
   DMA_ID_NEXT,
   DMA_ID_REF,
   DMA_ID_REFS,
   DMA_ID_CALL,
   DMA_ID_RET,
   DMA_ID_END,

   DMA_ID_MAX
};

typedef  struct
{
   unsigned int  qwc:16;
   unsigned int  rsvd:10;
   unsigned int  pce:2;
   unsigned int  id:3;
   unsigned int  irq:1;
   unsigned int  addr:31;
   unsigned int  spr:1;

   unsigned int  vt[2];
} DMA_Tag;

typedef struct
{
   unsigned int  Channel;        // Specify which channel the DMA chain is meant for

   unsigned int  Flip;           // 0 or 1, depending on the which of the 2 buffers is used

   unsigned int  BufferSize;     // Size of each of the 2 buffers
   unsigned int  SpaceLeft;      // Space left in the current buffer

   DMA_Tag       *Buffers[2];    // 2 buffers for double-buffering
   DMA_Tag       *CurrentPtr;    // End pointer in the current buffer

   unsigned int  Pad;            // To be 128-bit aligned
} DMA_ChainHeader;

//-------------------------------------
static VPUFILE  *DMA_vfd;
static ps2_vpu  *DMA_vpu[2];

//-------------------------------------
// Initialize the module
//-------------------------------------
unsigned int DMA_ModuleInit (VPUFILE *vfd, ps2_vpu *vpu0, ps2_vpu *vpu1, struct OS_Data *osdata)
{
   DMA_vfd = vfd;
   DMA_vpu[0] = vpu0;
   DMA_vpu[1] = vpu1;

   // Everything was fine
   return (DMA_RET_OK);
}

//-------------------------------------
// Uninitialize the module
//-------------------------------------
unsigned int DMA_ModuleUninit (void)
{
   // Currently nothing to do here
   return (DMA_RET_OK);
}

//-------------------------------------------------------------------------
// Initialize a double-buffered DMA chain and readies it to accept packets.
//-------------------------------------------------------------------------
unsigned int DMA_ChainInit (OS_Qword *chainbuffer, unsigned int buffersize,
                            unsigned int channel)
{
   DMA_ChainHeader  *header;
   unsigned int     channelnbr;

   // Make sure parameters are valid
   if (!DMA_POWEROF2 (channel))
      return (DMA_RET_INVALIDPARAMETER);
   for (channelnbr = 0; ((1 << channelnbr) != channel) && ((1 << channelnbr) != DMA_CHANNEL_MAX); channelnbr++) {}
   if ((chainbuffer == NULL) || (buffersize < sizeof (DMA_ChainHeader)) || (channelnbr >= 2))
      return (DMA_RET_INVALIDPARAMETER);

   // A DMA buffer must be 128-bit aligned (qword, 16 bytes)
   if ((unsigned int)chainbuffer & 0x0000000F)
      return (DMA_RET_WRONGALIGNMENT);

   // Fill out the DMA chain header
   {
      header = (DMA_ChainHeader *)chainbuffer;
      header->Channel = channelnbr;
      header->Flip = 0;

      header->BufferSize = ((buffersize - sizeof (DMA_ChainHeader)) & 0xFFFFFFF0) / 2;
      header->SpaceLeft = header->BufferSize;

      header->Buffers[0] = (DMA_Tag *)(header + 1);
      header->Buffers[1] = (DMA_Tag *)((unsigned int)header->Buffers[0] + header->BufferSize);
      header->CurrentPtr = header->Buffers[header->Flip];

      // Put an END tag
      header->CurrentPtr->qwc = 0;
      header->CurrentPtr->pce = 0;
      header->CurrentPtr->id = DMA_ID_END;
      header->CurrentPtr->irq = 0;
      header->CurrentPtr->addr = 0;
      header->CurrentPtr->spr = 0;
      header->CurrentPtr->vt[0] = 0;
      header->CurrentPtr->vt[1] = 0;
   }

   // Everything was fine
   return (DMA_RET_OK);
}

//-------------------------------------
// Initiate the DMA chain transfer to
// the associated channel
//-------------------------------------
unsigned int DMA_ChainSend (OS_Qword *chainbuffer)
{
   DMA_ChainHeader  *header;

   // Make sure parameters are valid
   if (chainbuffer == NULL)
      return (DMA_RET_INVALIDPARAMETER);

   header = (DMA_ChainHeader *)chainbuffer;

   // Disregard sending empty buffers
   if (header->Buffers[header->Flip] == header->CurrentPtr)
      return (DMA_RET_OK);

   ps2_dma_wait (ps2_vpu_fd (DMA_vpu[header->Channel]), 1);
   ps2_dma_start (ps2_vpu_fd (DMA_vpu[header->Channel]), DMA_vfd, (ps2_dmatag *)header->Buffers[header->Flip]);

   // Flip buffer
   header->Flip = 1 - header->Flip;
   header->CurrentPtr = header->Buffers[header->Flip];
   header->SpaceLeft = header->BufferSize;

   // Put an END tag
   header->CurrentPtr->qwc = 0;
   header->CurrentPtr->pce = 0;
   header->CurrentPtr->id = DMA_ID_END;
   header->CurrentPtr->irq = 0;
   header->CurrentPtr->addr = 0;
   header->CurrentPtr->spr = 0;
   header->CurrentPtr->vt[0] = 0;
   header->CurrentPtr->vt[1] = 0;

   // Everything was fine
   return (DMA_RET_OK);
}

//----------------------------------------------------------
// Wait until the DMA chain completely transfered
//----------------------------------------------------------
unsigned int DMA_ChainTransferWait (OS_Qword *chainbuffer)
{
   DMA_ChainHeader  *header;

   // Make sure parameters are valid
   if (chainbuffer == NULL)
      return (DMA_RET_INVALIDPARAMETER);

   header = (DMA_ChainHeader *)chainbuffer;
   
   ps2_dma_wait (ps2_vpu_fd (DMA_vpu[header->Channel]), 1);

   // Everything was fine
   return (DMA_RET_OK);
}

//----------------------------------------------------------
// Add data to the DMA chain, either by reference, inlining
// or call.  Reference is for large and static data (like
// geometry or textures), inlining is for small and changing
// data (like matrices or rendering states), and call is for
// pre-built DMA lists.
//----------------------------------------------------------
unsigned int DMA_ChainAdd (OS_Qword *chainbuffer, OS_Qword *buffer,
                           unsigned int buffersizeinqw, unsigned int attachment)
{
   unsigned int     requiredspace;
   DMA_ChainHeader  *header;

   // Make sure parameters are valid
   if ((chainbuffer == NULL) || (buffer == NULL) || ((unsigned int)buffer & 0xF) ||
       (attachment >= DMA_ATTACHMENT_MAX))
      return (DMA_RET_INVALIDPARAMETER);

   // A DMA buffer must be 128-bit aligned (qword, 16 bytes)
   if ((unsigned int)buffer & 0x0000000F)
      return (DMA_RET_WRONGALIGNMENT);

   header = (DMA_ChainHeader *)chainbuffer;

   switch (attachment)
   {
      case DMA_ATTACHMENT_INLINE:
            // We'll add a CNT tag, followed by the data, and then an END tag
            requiredspace = buffersizeinqw * sizeof (OS_Qword) + sizeof (DMA_Tag) + sizeof (DMA_Tag);

            // Make sure the buffer is big enough to fit more data
            if (requiredspace > header->SpaceLeft)
               return (DMA_RET_DATATOOBIGTOFIT);

            // The CNT tag
            header->CurrentPtr->qwc = buffersizeinqw;
            header->CurrentPtr->pce = 0;
            header->CurrentPtr->id = DMA_ID_CNT;
            header->CurrentPtr->irq = 0;
            header->CurrentPtr->addr = 0;
            header->CurrentPtr->spr = 0;
            header->CurrentPtr->vt[0] = 0;
            header->CurrentPtr->vt[1] = 0;
            header->CurrentPtr++;

            // The data itself
            memcpy (header->CurrentPtr, buffer, buffersizeinqw * sizeof (OS_Qword));
            header->CurrentPtr += buffersizeinqw;

            // The END tag
            header->CurrentPtr->qwc = 0;
            header->CurrentPtr->pce = 0;
            header->CurrentPtr->id = DMA_ID_END;
            header->CurrentPtr->irq = 0;
            header->CurrentPtr->addr = 0;
            header->CurrentPtr->spr = 0;
            header->CurrentPtr->vt[0] = 0;
            header->CurrentPtr->vt[1] = 0;

            // Notice that at this point, CurrentPtr points to
            // the END tag.  So if more data is added, it will
            // overwrite the now-unnecessary END tag.
            header->SpaceLeft -= requiredspace - sizeof (DMA_Tag);

            break;

      case DMA_ATTACHMENT_REFERENCE:
            // We'll add a REF tag, followed by an END tag
            requiredspace = sizeof (DMA_Tag) + sizeof (DMA_Tag);

            // Make sure the buffer is big enough to fit more data
            if (requiredspace > header->SpaceLeft)
               return (DMA_RET_DATATOOBIGTOFIT);

            // The REF tag
            header->CurrentPtr->qwc = buffersizeinqw;
            header->CurrentPtr->pce = 0;
            header->CurrentPtr->id = DMA_ID_REF;
            header->CurrentPtr->irq = 0;
            header->CurrentPtr->addr = (unsigned int)buffer;
            header->CurrentPtr->spr = 0;
            header->CurrentPtr->vt[0] = 0;
            header->CurrentPtr->vt[1] = 0;
            header->CurrentPtr++;

            // The END tag
            header->CurrentPtr->qwc = 0;
            header->CurrentPtr->pce = 0;
            header->CurrentPtr->id = DMA_ID_END;
            header->CurrentPtr->irq = 0;
            header->CurrentPtr->addr = 0;
            header->CurrentPtr->spr = 0;
            header->CurrentPtr->vt[0] = 0;
            header->CurrentPtr->vt[1] = 0;

            // Notice that at this point, CurrentPtr points to
            // the END tag.  So if more data is added, it will
            // overwrite the now-unnecessary END tag.
            header->SpaceLeft -= sizeof (DMA_Tag);

            break;

      case DMA_ATTACHMENT_CALL:
            // Not supported yet...  Sorry!
            return (DMA_RET_INVALIDPARAMETER);
            break;
   }

   // Everything was fine
   return (DMA_RET_OK);
}

//------------------------------------------------
// Return the size of the current chain, in qwords
//------------------------------------------------
unsigned int DMA_GetChainSize (OS_Qword *chainbuffer, unsigned int *chainsize)
{
   DMA_ChainHeader  *header;

   // Make sure parameters are valid
   if ((chainbuffer == NULL) || (chainsize == NULL))
      return (DMA_RET_INVALIDPARAMETER);

   header = (DMA_ChainHeader *)chainbuffer;

   *chainsize = ((unsigned int)header->CurrentPtr - (unsigned int)header->Buffers[header->Flip]) >> 4;

   // Everything was fine
   return (DMA_RET_OK);
}
